from drepr.executors.readers.geotiff import GeoTIFFReader
from drepr.executors.readers.netcdf import NetCDF4Reader
from drepr.executors.readers.np_dict import NPDictReader
from drepr.executors.readers.reader_container import ReaderContainer
from drepr.executors.readers.shapefile import ShapefileReader
from drepr.models import DRepr, ResourceType, RangeAlignment, PMap
import numpy as np

from drepr.executors.preprocessing.context import Context
from drepr.executors.preprocessing.py_exec import PyExec
from drepr.executors.readers.netcdf import NetCDF4Reader
from drepr.models import DRepr, ResourceType, PreprocessingType, IndexExpr, ClassNode, DataNode, LiteralNode, RangeExpr
from drepr.outputs.array_backend.array_attr import ArrayAttr, ScalarAttr, NoData


class CFConventionNDArrayMap:

    @classmethod
    def analyze(cls, ds_model: DRepr) -> bool:
        # only one resource
        if len(ds_model.resources) > 1:
            return False

        # ensure it is geotiff, netcdf
        if ds_model.resources[0].type not in {ResourceType.GeoTIFF, ResourceType.NetCDF3, ResourceType.NetCDF4, ResourceType.NPDict}:
            return False

        # only have dimension alignment
        if not all(isinstance(align, RangeAlignment) for align in ds_model.aligns):
            return False

        # only have map preprocessing, which mutate the current data
        for prepro in ds_model.preprocessing:
            if not isinstance(prepro.value, PMap) \
                    or prepro.value.output is not None \
                    or prepro.value.change_structure:
                return False

        return True

    @classmethod
    def execute(cls, ds_model: DRepr, resource_file: str):
        """
        Execution steps:

        1. create resource reader (only support one resource)
        2. execute preprocessing function (which is map function and only mutate the value)
        3. gather attributes
        4. Create table:
            a. If there are duplicated predicates (duplicated column_id), then the resulted array will have an extra dimension at
               the end, which size is equal to the number of duplicated columns of `column_id`. (see the ndarray documentation)
        """
        # 1st: create resource reader
        resource = ds_model.resources[0]
        if resource.type == ResourceType.NetCDF4:
            reader = NetCDF4Reader.from_file(resource_file)
        elif resource.type == ResourceType.GeoTIFF:
            reader = GeoTIFFReader.from_file(resource_file)
        elif resource.type == ResourceType.NPDict:
            reader = NPDictReader.from_file(resource_file)
        elif resource.type == ResourceType.Shapefile:
            reader = ShapefileReader.from_file(resource_file)
        elif resource.type == ResourceType.Container:
            reader = ReaderContainer.get_instance().get(resource_file)
        else:
            raise NotImplementedError()

        # 2nd: execute preprocessing function
        context = Context(reader)
        for preprocess_fn in ds_model.preprocessing:
            if preprocess_fn.type == PreprocessingType.pmap:
                steps = preprocess_fn.value.path.steps
                n_range_exprs = sum(int(not isinstance(step, IndexExpr)) for step in steps)
                fn = PyExec.compile(preprocess_fn.value.code)

                if n_range_exprs == 0:
                    index = [step.val for step in steps]
                    value = fn.exec(reader.get_value(index), index, context)
                    reader.set_value(index, value)
                else:
                    # TODO: improve it
                    range_indices = [i for i in range(len(steps)) if isinstance(steps[i], RangeExpr)]
                    index = [step.start if isinstance(step, RangeExpr) else step.val for step in steps]
                    if n_range_exprs == 1:
                        range_idx = range_indices[0]
                        if steps[range_idx].end is None:
                            if range_idx == 0:
                                end = reader.len_range()
                            else:
                                end = len(reader.get_value(index[:range_idx]))
                        else:
                            end = steps[range_idx].end
                        for i in range(steps[range_idx].start, end, steps[range_idx].step):
                            index[range_idx] = i
                            value = fn.exec(reader.get_value(index), index, context)
                            reader.set_value(index, value)
                    else:
                        raise NotImplementedError()
            else:
                raise Exception("You found a bug")

        # 3rd: gather attributes
        attrs = {}
        new_attrs = {}
        for attr in ds_model.attrs:
            attrs[attr.id] = reader.select(attr.path.steps)
            is_array_attr = any(not isinstance(step, IndexExpr) for step in attr.path.steps)
            if is_array_attr:
                step2dim = []
                path = attr.path
                if len(attr.missing_values) == 0:
                    nodata = None
                elif len(attr.missing_values) == 1:
                    # we need to convert this into numpy because comparing between numpy number
                    # and python number leads to error
                    if isinstance(attrs[attr.id], np.ndarray):
                        nodata = NoData(np.array(attr.missing_values, dtype=attrs[attr.id].dtype)[0])
                    else:
                        nodata = NoData(attr.missing_values[0])
                else:
                    # need to convert other values back to just one value and use it!
                    raise NotImplementedError()

                count = 0
                for step in path.steps:
                    if not isinstance(step, IndexExpr):
                        # in this function all the shape of ndarray will be determined by range only,
                        # and they will follow the order. Will we have other cases?
                        assert isinstance(step, RangeExpr)
                        step2dim.append(count)
                        count += 1
                    else:
                        step2dim.append(None)

                new_attrs[f"dnode:{attr.id}"] = ArrayAttr(f"dnode:{attr.id}", attrs[attr.id], path, step2dim, nodata)
            else:
                new_attrs[f"dnode:{attr.id}"] = ScalarAttr(f"dnode:{attr.id}", attrs[attr.id])

        # 4th: create tables from the semantic model
        sm = ds_model.sm
        tables = {}
        alignments = {}

        for nid, node in sm.nodes.items():
            if isinstance(node, ClassNode):
                outgoing_edges = list(sm.iter_outgoing_edges(nid))
                tables[nid] = {}
                for e in outgoing_edges:
                    c = sm.nodes[e.target_id]
                    if isinstance(c, DataNode):
                        if isinstance(attrs[c.attr_id], np.ndarray):
                            step2dim = []
                            _attr = ds_model.get_attr_by_id(c.attr_id)
                            path = _attr.path
                            if len(_attr.missing_values) == 0:
                                nodata = None
                            elif len(_attr.missing_values) == 1:
                                nodata = NoData(_attr.missing_values[0])
                            else:
                                # need to convert other values back to just one value and use it!
                                raise NotImplementedError()

                            count = 0
                            for step in path.steps:
                                if not isinstance(step, IndexExpr):
                                    # in this function all the shape of ndarray will be determined by range only,
                                    # and they will follow the order. Will we have other cases?
                                    assert isinstance(step, RangeExpr)
                                    step2dim.append(count)
                                    count += 1
                                else:
                                    step2dim.append(None)
                            # tables[nid][e.target_id] = ColArray(e.target_id, attrs[c.attr_id], path, step2dim, nodata)
                        else:
                            pass
                            # tables[nid][e.target_id] = ColSingle(attrs[c.attr_id])
                    elif isinstance(c, LiteralNode):
                        pass
                        # tables[nid][e.target_id] = ColSingle(c.value)
                    elif not isinstance(c, ClassNode):
                        raise NotImplementedError()
            elif isinstance(node, LiteralNode):
                new_attrs[nid] = ScalarAttr(nid, node.value)

        # inferring alignments
        for align in ds_model.aligns:
            source = f"dnode:" + align.source
            target = f"dnode:" + align.target
            if source not in alignments:
                alignments[source] = {}
            alignments[source][target] = align
        return None, new_attrs
        # return ndarray.NDArrayGraph(sm, tables, alignments), new_attrs



from typing import List, Dict, Tuple, Callable, Any, Optional, Union


class BlankRecordID(tuple):
    def __new__(cls, index: Tuple[int, ...], class_id: str):
        return super().__new__(cls, index)

    def __init__(self, index: Tuple[int, ...], class_id: str):
        self.index = index
        self.class_id = class_id


class URIRecordID(str):
    def __new__(cls, uri: str, index: Any, class_id: str):
        return super().__new__(cls, uri)

    def __init__(self, uri: str, index: Any, class_id: str):
        self.index = index
        self.class_id = class_id


class GroupRecordID(str):
    def __new__(cls, uri: str, indice: List[Any], class_id: str):
        return super().__new__(cls, uri)

    def __init__(self, uri: str, indice: List[Any], class_id: str):
        self.indice = indice
        self.class_id = class_id


class GraphRecordID(str):
    def __new__(cls, id: str, class_id: str) -> Any:
        return super().__new__(cls, id)

    def __init__(self, id: str, class_id: str):
        self.class_id = class_id

    def __copy__(self):
        return GraphRecordID(self, self.class_id)

    def __deepcopy__(self, memodict={}):
        return GraphRecordID(self, self.class_id)


ArrayRecordID = Union[BlankRecordID, URIRecordID, GroupRecordID]
RecordID = Union[GraphRecordID, ArrayRecordID]


if __name__ == '__main__':
    rid = URIRecordID("https://example.org", (1, 2, 3))
    print(rid)
    print(rid.index)


from dataclasses import dataclass
from enum import Enum
from typing import Union, Optional


class ResourceType(Enum):
    CSV = "csv"
    JSON = "json"
    XML = "xml"
    Spreadsheet = "spreadsheet"
    NetCDF4 = "netcdf4"
    NetCDF3 = "netcdf3"
    GeoTIFF = "geotiff"
    NPDict = "np-dict"
    Shapefile = "shapefile"
    Container = "container"


@dataclass
class CSVProp:
    delimiter: str = ","


@dataclass
class Resource:
    id: str
    type: ResourceType
    prop: Optional[Union[CSVProp]] = None

    @staticmethod
    def deserialize(raw: dict):
        if raw['type'] == ResourceType.CSV.value and raw['prop'] is not None:
            prop = CSVProp(raw['prop']['delimiter'])
        else:
            prop = None
        return Resource(raw['id'], ResourceType(raw['type']), prop)



from typing import List, Dict, Tuple, Callable, Any, Optional

from pathlib import Path

import pytest

from drepr.outputs.array_backend.array_backend import ArrayBackend
from drepr.outputs.graph_backend.graph_backend import GraphBackend


def get_backends(dataset_dir: Path):
    return [
        ArrayBackend.from_drepr(str(dataset_dir / "model.yml"), str(dataset_dir / "resource.json")),
        GraphBackend.from_drepr(str(dataset_dir / "model.yml"), str(dataset_dir / "resource.json"))
    ]


@pytest.fixture()
def s01(resource_dir):
    return get_backends(resource_dir / "s01_synthesis")


@pytest.fixture()
def s02(resource_dir):
    return get_backends(resource_dir / "s02_synthesis")


@pytest.fixture()
def s03(resource_dir):
    return get_backends(resource_dir / "s03_synthesis")


@pytest.fixture()
def s05(resource_dir):
    return get_backends(resource_dir / "s05_crop_place")


import copy
from typing import Dict, Tuple

import ujson
from netCDF4 import Dataset

from drepr.models import ResourceType, DRepr, LiteralNode, Preprocessing, PreprocessingType, PMap, DataNode, Attr, Path, \
    IndexExpr, Sorted, Edge, Resource, ClassNode, RangeAlignment
from drepr.patches import ResourceData, ResourceDataFile, ResourceDataString


def patch(repr: DRepr, resources: Dict[str, ResourceData]) -> Tuple[DRepr, Dict[str, ResourceData]]:
    """
    This patch turns classes that have only static properties to have one attribute in order to keep the
    algorithm implemented in Rust simple and fast
    """
    patch_classes = []
    for u in repr.sm.iter_class_nodes():
        if all(isinstance(repr.sm.nodes[e.target_id], LiteralNode) for e in repr.sm.iter_outgoing_edges(u.node_id)):
            patch_classes.append(u)

    if len(patch_classes) > 0:
        repr = copy.deepcopy(repr)
        resources = {k: resources[k] for k in resources}

        resource_id = '__static_patch_resource__'
        assert resource_id not in resources
        resource_data = {}

        existing_attrs = {a.id for a in repr.attrs}
        for u in patch_classes:
            replace_edge = None
            for e in repr.sm.iter_outgoing_edges(u.node_id):
                if e.label == 'drepr:uri':
                    replace_edge = e
                    break
                replace_edge = e
            v = repr.sm.nodes[replace_edge.target_id]

            new_attr_id = f'a87k219da_{u.node_id}_{replace_edge.edge_id}_{v.node_id}'
            assert new_attr_id not in existing_attrs
            resource_data[new_attr_id] = v.value

            # remove literal node
            repr.sm.remove_node(replace_edge.target_id)
            # repr.sm.edges.pop(replace_edge.edge_id)

            # add new attribute and link to it
            repr.attrs.append(Attr(new_attr_id, resource_id, Path([IndexExpr(new_attr_id)]), [], unique=True, sorted=Sorted.Ascending))
            v = DataNode(f'dnode:{new_attr_id}', new_attr_id)
            repr.sm.nodes[v.node_id] = v
            repr.sm.edges[replace_edge.edge_id] = Edge(replace_edge.edge_id, replace_edge.source_id, v.node_id, replace_edge.label, is_subject=True)

            # add alignment with incoming class, add all attributes since we don't know which attribute will be the subject
            for e in repr.sm.iter_incoming_edges(u.node_id):
                if isinstance(repr.sm.nodes[e.source_id], ClassNode):
                    for ie in repr.sm.iter_outgoing_edges(e.source_id):
                        if isinstance(repr.sm.nodes[ie.target_id], DataNode):
                            repr.aligns.append(RangeAlignment(repr.sm.nodes[ie.target_id].attr_id, new_attr_id, []))
            for e in repr.sm.iter_outgoing_edges(u.node_id):
                if isinstance(repr.sm.nodes[e.target_id], ClassNode):
                    for ie in repr.sm.iter_outgoing_edges(e.target_id):
                        if isinstance(repr.sm.nodes[ie.target_id], DataNode):
                            repr.aligns.append(RangeAlignment(new_attr_id, repr.sm.nodes[ie.target_id].attr_id, []))

        repr.resources.append(Resource(resource_id, ResourceType.JSON))
        resources[resource_id] = ResourceDataString(ujson.dumps(resource_data))

    return repr, resources





from abc import ABC, abstractmethod
from typing import List, Dict, Tuple, Callable, Any, Optional, Iterable, Union

from drepr.outputs.record_id import RecordID
from drepr.outputs.base_output_class import BaseOutputClass
from drepr.outputs.base_output_predicate import BaseOutputPredicate
from drepr.outputs.base_record import BaseRecord


class BaseLstOutputClass(ABC):

    @abstractmethod
    def iter_records(self) -> Iterable[BaseRecord]:
        """
        """
        pass

    @abstractmethod
    def filter(self, conditions: Union['FCondition', List['FCondition']]) -> 'BaseLstOutputClass':
        pass

    @abstractmethod
    def group_by(self, predicate: Union[str, BaseOutputPredicate]) -> Iterable[Tuple[Any, 'BaseLstOutputClass']]:
        """Group records of this classes by values of a predicate"""
        pass

    @abstractmethod
    def __iter__(self) -> Iterable[BaseOutputClass]:
        pass

    @abstractmethod
    def __len__(self) -> int:
        pass

    @abstractmethod
    def __getitem__(self, item: int) -> BaseOutputClass:
        pass


from typing import List, Dict, Tuple, Callable, Any, Optional


class Namespace:
    RDF = "http://www.w3.org/1999/02/22-rdf-syntax-ns#"

    def __init__(self, base: str, seen_uris: Optional[Dict[str, str]] = None):
        self._base = base
        self.seen_uris = seen_uris or {}

    def __getattr__(self, item):
        if item not in self.seen_uris:
            uri = self._base + item
            self.seen_uris[item] = uri
        return self.seen_uris[item]

    def __getitem__(self, item):
        if item not in self.seen_uris:
            uri = self._base + item
            self.seen_uris[item] = uri
        return self.seen_uris[item]


class PrefixedNamespace(Namespace):

    def __init__(self, base: str, prefix: str, seen_uris: Optional[Dict[str, str]] = None):
        super().__init__(base, seen_uris)
        self._prefix = prefix + ":"

    def __getattr__(self, item):
        if item not in self.seen_uris:
            uri = self._prefix + item
            self.seen_uris[item] = uri
        return self.seen_uris[item]

    def __getitem__(self, item):
        if item not in self.seen_uris:
            uri = self._prefix + item
            self.seen_uris[item] = uri
        return self.seen_uris[item]



from typing import List

from drepr.outputs.array_backend.array_backend import ArrayBackend
from drepr.outputs.array_backend.array_class import ArrayClass
from drepr.outputs.array_backend.lst_array_class import LstArrayClass
from drepr.outputs.base_lst_output_class import BaseLstOutputClass
from drepr.outputs.base_output_class import BaseOutputClass

from drepr.outputs.base_output_sm import BaseOutputSM
from drepr.outputs.record_id import BlankRecordID, GraphRecordID


def test_from_drepr(s01, s02, s03, s05):
    """This function does a smoke test to ensure that we can output all datasets"""
    pass


def test_get_class_by_id(s01: List[BaseOutputSM], s02: List[BaseOutputSM]):
    # has single class and get them by id
    for sm in (item for lst in [s01, s02] for item in lst):
        cls_id = sm.c("mint:Variable")[0].id
        cls = sm.cid(cls_id)
        assert isinstance(cls, BaseOutputClass)
        assert cls.id == cls_id


def test_get_classes_by_uri(s01: List[BaseOutputSM], s03: List[BaseOutputSM]):
    # for single class, directly get array class
    for sm in s01:
        lst = sm.c("mint:Variable")
        assert isinstance(lst, BaseLstOutputClass) and len(lst) == 1 and lst[0].uri == "mint:Variable"

    for sm in s03:
        lst = sm.c("mint:Variable")
        assert isinstance(lst, BaseLstOutputClass) and len(lst) == 2
        assert all(x.uri == "mint:Variable" for x in lst)


def test_get_record_by_id_blank(s01: List[BaseOutputSM]):
    for sm in s01:
        if isinstance(sm, ArrayBackend):
            record = sm.get_record_by_id(BlankRecordID((), sm.c("mint-geo:Raster")[0].id))
        else:
            record = sm.get_record_by_id(GraphRecordID("_:Raster1", sm.c("mint-geo:Raster")[0].id))

        d = record.to_dict()
        d.pop('@id')
        assert d == {
            'mint-geo:dx': [0.1],
            'mint-geo:dy': [0.1],
            'mint-geo:epsg': [4326],
            'mint-geo:x_min': [44.1],
            'mint-geo:x_slope': [0.0],
            'mint-geo:y_min': [20.1],
            'mint-geo:y_slope': [0.0]
        }



from typing import List, Tuple, Any, Union, Iterable, TYPE_CHECKING

from drepr.outputs.base_output_predicate import BaseOutputPredicate
from drepr.outputs.base_record import BaseRecord
from drepr.outputs.base_subset_output_class import BaseSubsetOutputClass
from drepr.outputs.graph_backend.graph_record import GraphRecord
from drepr.outputs.record_id import RecordID

if TYPE_CHECKING:
    from drepr.outputs.graph_backend.graph_class import GraphClass


class SubsetGraphClass(BaseSubsetOutputClass):

    def __init__(self, cls: 'GraphClass', nodes: List[GraphRecord]) -> None:
        super().__init__()
        self.cls = cls
        self.nodes = nodes

    def is_blank(self) -> bool:
        return self.cls.is_blank()

    def iter_records(self) -> Iterable[GraphRecord]:
        return self.nodes

    def get_record_by_id(self, rid: RecordID) -> BaseRecord:
        pass

    def p(self, predicate_uri: str) -> List[BaseOutputPredicate]:
        return self.cls.p(predicate_uri)

    def o(self, predicate_uri: str, target_uri: str) -> List['BaseOutputClass']:
        # return self.cls.p()
        pass

    def filter(self, conditions: Union['FCondition', List['FCondition']]) -> 'BaseSubsetOutputClass':
        pass

    def group_by(self, predicate: Union[str, BaseOutputPredicate]) -> Iterable[Tuple[Any, 'BaseSubsetOutputClass']]:
        pass


from abc import ABC, abstractmethod
from typing import List, Dict, Tuple, Callable, Any, Optional, Union


class BaseRecord(ABC):
    id: Union[str, tuple]

    @abstractmethod
    def s(self, predicate_uri: str) -> Any:
        pass

    @abstractmethod
    def m(self, predicate_uri: str) -> Any:
        pass

    @abstractmethod
    def us(self, predicate_uri: str, val: Any):
        pass

    @abstractmethod
    def um(self, predicate_uri: str, val: Any):
        pass

    @abstractmethod
    def to_dict(self) -> dict:
        """Return a copied data of this record as a dictionary"""
        pass


from .array_backend.array_backend import ArrayBackend
from .graph_backend.graph_backend import GraphBackend
from .namespace import Namespace
from .base_output_sm import FCondition
from .prop_data_ndarray import PropDataNDArray


import copy
from collections import defaultdict
from typing import List, Dict, Tuple, Callable, Any, Optional

from drepr.models import DRepr, WildcardExpr, StepExpr, RangeExpr, Expr, IndexExpr, Preprocessing, PreprocessingType, \
    RMap, RMapFunc, Path, RangeAlignment
from drepr.patches import ResourceData


def patch(repr: DRepr, _resources: Dict[str, ResourceData]) -> DRepr:
    """
    This patch will fix json path that select property names, before we officially support
    selecting property names

    How does it work: it creates a preprocessing function that convert an object to list of key-value pairs
    """
    need_patch = False
    for pfunc in repr.preprocessing:
        for step in pfunc.value.path.steps:
            if step == WildcardExpr.Names:
                raise Exception(
                    "Does not support preprocessing that is applied on properties of an object yet!"
                )

    for attr in repr.attrs:
        n_wildcard_names = sum(int(step == WildcardExpr.Names) for step in attr.path.steps)
        if n_wildcard_names > 1:
            # cannot select names inside names
            raise Exception(f"Invalid path for attribute {attr.id}")

        if n_wildcard_names > 0:
            need_patch = True

    if need_patch:
        repr = copy.deepcopy(repr)

        # first, find paths that containing the WildcardExpr.Names (*~)
        property_paths = []
        for attr in repr.attrs:
            for i, step in enumerate(attr.path.steps):
                if step == WildcardExpr.Names:
                    property_paths.append((attr.id, i, attr.path.steps[:i]))
                    break

        # now we need to make sure for every property paths, all paths that is overlap to the property paths,
        # need to be exactly equal and no property paths that is subpath of another property path
        for i in range(len(property_paths)):
            for j in range(i + 1, len(property_paths)):
                if is_path_overlap(property_paths[i][-1], property_paths[j][-1]):
                    raise Exception(
                        f"Does not handle nested selecting property names yet. Found in either {property_paths[i][0]} or {property_paths[j][0]}"
                    )

        overlapped_paths = defaultdict(lambda: [])
        for attr in repr.attrs:
            steps = attr.path.steps
            for i, (paid, _, path) in enumerate(property_paths):
                if len(steps) > len(path) and is_path_overlap(path, steps):
                    assert steps[:len(path)] == path, f"Slices {steps} does not fully contain {path}"
                    overlapped_paths[i].append(attr.id)
        # add preprocessing functions
        id2attr = {attr.id: attr for attr in repr.attrs}
        for paid, _, path in property_paths:
            # need to convert dict to key-value pairs
            repr.preprocessing.append(Preprocessing(
                PreprocessingType.rmap,
                RMap(id2attr[paid].resource_id, Path(path), RMapFunc.Dict2Items)
            ))

        # now every property path is distinct, and we have a map of property path => attributes
        # that require to re-map their positions
        for pi, aids in overlapped_paths.items():
            _, _, path = property_paths[pi]
            for aid in aids:
                # remap attribute slices
                attr = id2attr[aid]
                steps = attr.path.steps

                if steps[len(path)] == WildcardExpr.Names:
                    # this select property names
                    steps = steps[:len(path)] + [RangeExpr(0, None, 1), IndexExpr(0)] + steps[len(path) + 1:]
                else:
                    assert steps[len(
                        path
                    )] == WildcardExpr.Values, 'Does not support select both all property names and just a subset of its values'
                    steps = steps[:len(path)] + [RangeExpr(0, None, 1), IndexExpr(1)] + steps[len(path) + 1:]
                attr.path.steps = steps

                # remap alignments
                for alignment in repr.aligns:
                    if isinstance(alignment, RangeAlignment):
                        if alignment.source == aid:
                            for adim in alignment.aligned_steps:
                                if adim.source_idx > len(path):
                                    adim.source_idx += 1
                        elif alignment.target == aid:
                            for adim in alignment.aligned_steps:
                                if adim.target_idx > len(path):
                                    adim.target_idx += 1

    return repr


def is_path_overlap(path_0: List[StepExpr], path_1: List[StepExpr]):
    def is_slice_overlap(s0, s1):
        if isinstance(s0, RangeExpr):
            if any(isinstance(v, Expr) for v in [s0.start, s0.end, s0.step]):
                raise Exception("Haven't support compare expr overlapped yet")

            if isinstance(s1, RangeExpr):
                if any(isinstance(v, Expr) for v in [s1.start, s1.end, s1.step]):
                    raise Exception("Haven't support compare expr overlapped yet")

                return s1.start < (s0.end or float('inf')) and (s1.end or float('inf')) > s0.start
            elif isinstance(s1, IndexExpr):
                if isinstance(s1.val, str) or isinstance(s1.val, Expr):
                    return False
                elif s0.start <= s1.val < (s0.end or float('inf')):
                    return True
                return False
        elif isinstance(s0, IndexExpr):
            if isinstance(s1, RangeExpr):
                if any(isinstance(v, Expr) for v in [s1.start, s1.end, s1.step]):
                    raise Exception("Haven't support compare expr overlapped yet")

                if isinstance(s0.val, (str, Expr)):
                    return False
                elif s1.start <= s0.val < (s1.end or float('inf')):
                    return True
                return False
            elif isinstance(s1, IndexExpr):
                return s1.val == s1.val
        else:
            # not index & range does not have overlapping?
            return s0 == s1

    if len(path_0) > len(path_1):
        path_0, path_1 = path_1, path_0

    for i in range(len(path_0)):
        s0 = path_0[i]
        s1 = path_1[i]

        if not is_slice_overlap(s0, s1):
            return False

    return True



from typing import List, Dict, Tuple, Callable, Any, Optional, Union, Iterable


class O2ORange0Func:
    """
    A function that map index of the primary attribute's item to target attribute's item.
    The mapping function is always *-to-1 mapping.
    """
    def __init__(self, target2source: List[int]):
        self.target2source = target2source
        self.is_x2o = True

    def __call__(self, source_idx: Tuple[int, ...]) -> Tuple[int, ...]:
        return tuple(source_idx[i] for i in self.target2source)


class O2MRange0Func:
    """
    """
    def __init__(self):
        self.is_x2o = False

    def __call__(self, source_idx: Tuple[int, ...]) -> Iterable[Tuple[int, ...]]:
        pass


class IdentityFunc:

    def __init__(self):
        self.is_x2o = True

    def __call__(self, source_idx: Tuple[int, ...]) -> Tuple[int, ...]:
        return source_idx


X2OFunc = Union[IdentityFunc, O2ORange0Func]
X2MFunc = Union[O2MRange0Func]



from typing import List, Dict, Tuple, Callable, Any, Optional, Iterable, TYPE_CHECKING

from drepr.outputs.array_backend.array_record import ArrayRecord
from drepr.outputs.base_output_predicate import BaseOutputPredicate
from drepr.outputs.base_subset_output_class import BaseSubsetOutputClass

if TYPE_CHECKING:
    from drepr.outputs.array_backend.array_class import ArrayClass
from drepr.outputs.record_id import RecordID


class SubsetArrayClass(BaseSubsetOutputClass):
    def is_blank(self) -> bool:
        pass

    def p(self, predicate_uri: str) -> List[BaseOutputPredicate]:
        pass

    def o(self, predicate_uri: str, target_uri: str) -> List['BaseSubsetOutputClass']:
        pass

    def __init__(self, cls: 'ArrayClass', ids: Iterable[RecordID]):
        self.cls = cls
        self.ids = ids

    def iter_records(self):
        for id in self.ids:
            yield ArrayRecord(id.index, self.cls)

    def get_record_by_id(self, rid: RecordID):
        return self.cls.get_record_by_id(rid)

    def get_data_prop_as_ndarray(self, edge_id: int, index_edge_ids: List[int]):
        pass

    def filter(self, conditions):
        pass

    def group_by(self, pred_uri: str):
        pass



from drepr import DRepr
from drepr.models.align import RangeAlignment, AlignedStep
from drepr.models.attr import Attr, Sorted, ValueType
from drepr.models.drepr import yaml
from drepr.models.parse_v1.path_parser import PathParser, PathParserV1
from drepr.models.preprocessing import Preprocessing, PMap, PreprocessingType
from drepr.models.resource import Resource, ResourceType, CSVProp
from drepr.models.sm import ClassNode, DataNode, SemanticModel, DataType, Edge

content = """
version: '1'
resources:
  default:
    type: csv
preprocessing:
  - type: pmap
    path: $[2:][1:]
    code: |
      return float(value)
attributes:
  area: 
    resource_id: default
    path: $[2:][0]
    missing_values: [-999]
    unique: true
    sorted: ascending
    value_type: list[int]
  gender: $[1][1:]
  period: $[0][1:]
  obs: $[2:][1:]
alignments:
  - type: dimension
    source: obs
    target: area
    aligned_dims:
      - { source: 0, target: 0 }
  - type: dimension
    value: obs:1 <-> gender:1
  - type: dimension
    value: obs:1 <-> period:1
semantic_model:
  data_nodes:
    area: qb:Observation:1--eg:refArea
    gender: qb:Observation:1--eg:gender
    period: qb:Observation:1--eg:refPeriod^^xsd:anyURI
    obs: qb:Observation:1--smdx-measure:obsValue
  relations: []
  prefixes:
    qb: http://purl.org/linked-data/cube#
    smdx-measure: http://purl.org/linked-data/sdmx/2009/measure#
    eg: http://example.org/
    """


def test_parse():
    ds_model = DRepr.parse(yaml.load(content))
    gold_resource = Resource("default", ResourceType.CSV, CSVProp(delimiter=","))
    gold_ds_model = DRepr(
        [gold_resource],
        [Preprocessing(PreprocessingType.pmap,
                       PMap("default", PathParserV1().parse(gold_resource, "$[2:][1:]", ""), "return float(value)\n", None, None))],
        [
            Attr("area", "default", PathParserV1().parse(gold_resource, "$[2:][0]", ""), [-999], True, Sorted.Ascending,
                 ValueType.List_Int),
            Attr("gender", "default", PathParserV1().parse(gold_resource, "$[1][1:]", ""), [], False, Sorted.Null, ValueType.Unspecified),
            Attr("period", "default", PathParserV1().parse(gold_resource, "$[0][1:]", ""), [], False, Sorted.Null, ValueType.Unspecified),
            Attr("obs", "default", PathParserV1().parse(gold_resource, "$[2:][1:]", ""), [], False, Sorted.Null,
                 ValueType.Unspecified),
        ],
        [
            RangeAlignment('obs', 'area', [AlignedStep(0, 0)]),
            RangeAlignment('obs', 'gender', [AlignedStep(1, 1)]),
            RangeAlignment('obs', 'period', [AlignedStep(1, 1)])
        ],
        SemanticModel(
            nodes={
                'qb:Observation:1': ClassNode('qb:Observation:1', 'qb:Observation'),
                'dnode:area': DataNode('dnode:area', 'area'),
                'dnode:gender': DataNode('dnode:gender', 'gender'),
                'dnode:period': DataNode('dnode:period', 'period', DataType('xsd:anyURI')),
                'dnode:obs': DataNode('dnode:obs', 'obs')
            },
            edges={
                0: Edge(0, 'qb:Observation:1',
                     'dnode:area',
                     'eg:refArea',
                     ),
                1: Edge(1, 'qb:Observation:1',
                     'dnode:gender',
                     'eg:gender',
                     ),
                2: Edge(2, 'qb:Observation:1',
                     'dnode:period',
                     'eg:refPeriod',
                     ),
                3: Edge(3, 'qb:Observation:1',
                     'dnode:obs',
                     'smdx-measure:obsValue',
                     )
            },
            prefixes={
                'qb': 'http://purl.org/linked-data/cube#',
                'smdx-measure': 'http://purl.org/linked-data/sdmx/2009/measure#',
                'eg': 'http://example.org/',
                'drepr': 'https://purl.org/drepr/1.0/',
                'rdf': 'http://www.w3.org/1999/02/22-rdf-syntax-ns#',
                'rdfs': 'http://www.w3.org/2000/01/rdf-schema#',
                'owl': 'http://www.w3.org/2002/07/owl#'

            }
        )
    )

    assert ds_model == gold_ds_model


def test_serialize():
    ds_model = DRepr.parse(yaml.load(content))
    assert ds_model.serialize() == {
        'resources': [{'id': 'default', 'prop': {'delimiter': ','}, 'type': 'csv'}],
        'preprocessing': [{
            'type': 'pmap',
            'value': {
                'change_structure': None,
                'code': 'return float(value)\n',
                'output': None,
                'path': {
                    'steps': [
                        {'end': None, 'start': 2, 'step': 1},
                        {'end': None, 'start': 1, 'step': 1}
                    ]}
                ,
                'resource_id': 'default'}
        }],
        'attrs': [
            {
                'id': 'area',
                'missing_values': [-999],
                'path': {
                    'steps': [
                        {'end': None, 'start': 2, 'step': 1},
                        {'val': 0}
                    ]
                },
                'resource_id': 'default',
                'sorted': 'ascending',
                'unique': True,
                'value_type': 'list[int]'
            },
            {
                'id': 'gender',
                'missing_values': [],
                'path': {
                    'steps': [
                        {'val': 1},
                        {'end': None, 'start': 1, 'step': 1}
                    ]
                },
                'resource_id': 'default',
                'sorted': 'none',
                'unique': False,
                'value_type': 'unspecified'},
            {
                'id': 'period',
                'missing_values': [],
                'path': {
                    'steps': [
                        {'val': 0},
                        {'end': None, 'start': 1, 'step': 1}
                    ]
                },
                'resource_id': 'default',
                'sorted': 'none',
                'unique': False,
                'value_type': 'unspecified'},
            {
                'id': 'obs',
                'missing_values': [],
                'path': {
                    'steps': [
                        {'end': None, 'start': 2, 'step': 1},
                        {'end': None, 'start': 1, 'step': 1}
                    ]
                },
                'resource_id': 'default',
                'sorted': 'none',
                'unique': False,
                'value_type': 'unspecified'
            }
        ],
        'aligns': [
            {
                'type': 'range',
                'aligned_steps': [{'source_idx': 0, 'target_idx': 0}],
                'source': 'obs',
                'target': 'area'
            },
            {
                'type': 'range',
                'aligned_steps': [{'source_idx': 1, 'target_idx': 1}],
                'source': 'obs',
                'target': 'gender'
            },
            {
                'type': 'range',
                'aligned_steps': [{'source_idx': 1, 'target_idx': 1}],
                'source': 'obs',
                'target': 'period'
            }
        ],
        'sm': {
            'edges': {
                0: {
                    'edge_id': 0,
                    'is_required': False,
                    'is_subject': False,
                    'label': 'eg:refArea',
                    'source_id': 'qb:Observation:1',
                    'target_id': 'dnode:area'
                },
                1: {
                    'edge_id': 1,
                    'is_required': False,
                    'is_subject': False,
                    'label': 'eg:gender',
                    'source_id': 'qb:Observation:1',
                    'target_id': 'dnode:gender'
                },
                2: {
                    'edge_id': 2,
                    'is_required': False,
                    'is_subject': False,
                    'label': 'eg:refPeriod',
                    'source_id': 'qb:Observation:1',
                    'target_id': 'dnode:period'
                },
                3: {
                    'edge_id': 3,
                    'is_required': False,
                    'is_subject': False,
                    'label': 'smdx-measure:obsValue',
                    'source_id': 'qb:Observation:1',
                    'target_id': 'dnode:obs'
                }
            },
            'nodes': {
                'dnode:area': {
                    'type': 'data_node',
                    'attr_id': 'area',
                    'data_type': None,
                    'node_id': 'dnode:area'
                },
                'dnode:gender': {
                    'type': 'data_node',
                    'attr_id': 'gender',
                    'data_type': None,
                    'node_id': 'dnode:gender'
                },
                'dnode:obs': {
                    'type': 'data_node',
                    'attr_id': 'obs',
                    'data_type': None,
                    'node_id': 'dnode:obs'
                },
                'dnode:period': {
                    'type': 'data_node',
                    'attr_id': 'period',
                    'data_type': 'xsd:anyURI',
                    'node_id': 'dnode:period'
                },
                'qb:Observation:1': {
                    'type': 'class_node',
                    'label': 'qb:Observation',
                    'node_id': 'qb:Observation:1'
                }
            },
            'prefixes': {
                'drepr': 'https://purl.org/drepr/1.0/',
                'eg': 'http://example.org/',
                'owl': 'http://www.w3.org/2002/07/owl#',
                'qb': 'http://purl.org/linked-data/cube#',
                'rdf': 'http://www.w3.org/1999/02/22-rdf-syntax-ns#',
                'rdfs': 'http://www.w3.org/2000/01/rdf-schema#',
                'smdx-measure': 'http://purl.org/linked-data/sdmx/2009/measure#'
            }
        }
    }


def test_deserialize():
    ds_model = DRepr.parse(yaml.load(content))
    assert ds_model == DRepr.deserialize(ds_model.serialize())


def test_write_to_yml():
    ds_model = DRepr.parse(yaml.load(content))
    assert ds_model.to_lang_yml(True) == """version: '2'
resources:
  default:
    type: csv
    delimiter: ','
preprocessing:
- type: pmap
  resource_id: default
  path:
  - 2..:1
  - 1..:1
  code: |
    return float(value)
  output:
  change_structure:
attributes:
  area:
    resource_id: default
    path:
    - 2..:1
    - 0
    unique: true
    sorted: ascending
    value_type: list[int]
    missing_values: [-999]
  gender:
    resource_id: default
    path:
    - 1
    - 1..:1
    unique: false
    sorted: none
    value_type: unspecified
    missing_values: []
  period:
    resource_id: default
    path:
    - 0
    - 1..:1
    unique: false
    sorted: none
    value_type: unspecified
    missing_values: []
  obs:
    resource_id: default
    path:
    - 2..:1
    - 1..:1
    unique: false
    sorted: none
    value_type: unspecified
    missing_values: []
alignments:
- type: range
  source: obs
  target: area
  aligned_dims:
  - source: 0
    target: 0
- type: range
  source: obs
  target: gender
  aligned_dims:
  - source: 1
    target: 1
- type: range
  source: obs
  target: period
  aligned_dims:
  - source: 1
    target: 1
semantic_model:
  qb:Observation:1:
    properties:
    - - eg:refArea
      - area
    - - eg:gender
      - gender
    - - eg:refPeriod
      - period
      - xsd:anyURI
    - - smdx-measure:obsValue
      - obs
    static_properties: []
    links: []
  prefixes:
    qb: http://purl.org/linked-data/cube#
    smdx-measure: http://purl.org/linked-data/sdmx/2009/measure#
    eg: http://example.org/
    drepr: https://purl.org/drepr/1.0/
    rdf: http://www.w3.org/1999/02/22-rdf-syntax-ns#
    rdfs: http://www.w3.org/2000/01/rdf-schema#
    owl: http://www.w3.org/2002/07/owl#
"""
    ds_model2 = DRepr.parse(yaml.load(ds_model.to_lang_yml(True)))
    assert ds_model2 == ds_model


from collections import defaultdict
from dataclasses import asdict
from typing import List

from drepr.models.parse_v2.path_parser import PathParserV2
from drepr.utils.validator import *
from ..align import AlignmentType, RangeAlignment

from ..parse_v1.align_parser import AlignParser
from ..parse_v1.attr_parser import AttrParser
from ..parse_v1.preprocessing_parser import PreprocessingParser
from ..parse_v1.resource_parser import ResourceParser

from .sm_parser import SMParser
from ..sm import SemanticModel, ClassNode, DataNode, LiteralNode


class ReprV2Parser:
    """
    The D-REPR language version 2 has similar to the schema of the first version.

    Difference with previous features:
    1. For spreadsheet columns, they can the letter instead of number
    2. Semantic model configuration is changed to focus on classes
    """

    TOP_KEYWORDS = {
        "version", "resources", "preprocessing", "attributes", "alignments", "semantic_model"
    }
    DEFAULT_RESOURCE_ID = "default"

    @classmethod
    def parse(cls, raw: dict):
        from ..drepr import DRepr

        Validator.must_be_subset(cls.TOP_KEYWORDS,
                                 raw.keys(),
                                 setname="Keys of D-REPR configuration",
                                 error_msg="Parsing D-REPR configuration")

        for prop in ['version', 'resources', 'attributes']:
            Validator.must_have(raw, prop, error_msg="Parsing D-REPR configuration")

        Validator.must_equal(raw['version'], '2', "Parsing D-REPR configuration version")
        resources = ResourceParser.parse(raw['resources'])

        if len(resources) == 1:
            default_resource_id = resources[0].id
        else:
            default_resource_id = ResourceParser.DEFAULT_RESOURCE_ID

        path_parser = PathParserV2()
        preprocessing = PreprocessingParser(path_parser).parse(default_resource_id, resources,
                                                               raw.get('preprocessing', []))
        attrs = AttrParser(path_parser).parse(default_resource_id, resources, raw['attributes'])
        aligns = AlignParser.parse(raw.get('alignments', []))

        if 'semantic_model' in raw:
            sm = SMParser.parse(raw['semantic_model'])
            sm.prefixes.update(SemanticModel.get_default_prefixes())
        else:
            sm = None

        return DRepr(resources, preprocessing, attrs, aligns, sm)

    @classmethod
    def dump(cls, drepr: 'DRepr', simplify: bool = True, use_json_path: bool = False):
        version = '2'
        sm = OrderedDict()

        class_counter = defaultdict(int)
        class_ids: Dict[str, str] = {}
        for node in drepr.sm.nodes.values():
            if isinstance(node, ClassNode):
                class_counter[node.label] += 1
                class_ids[node.node_id] = f"{node.label}:{class_counter[node.label]}"
                sm[class_ids[node.node_id]] = OrderedDict([
                    ("properties", []),
                    ("static_properties", []),
                    ("links", [])
                ])

        for node in drepr.sm.nodes.values():
            if isinstance(node, DataNode):
                edge = [e for e in drepr.sm.edges.values() if e.target_id == node.node_id][0]
                if node.data_type is not None:
                    prop = (edge.label, node.attr_id, node.data_type.value)
                else:
                    prop = (edge.label, node.attr_id)
                sm[class_ids[edge.source_id]]['properties'].append(prop)

            if isinstance(node, LiteralNode):
                edge = [e for e in drepr.sm.edges if e.target_id == node.node_id][0]
                if node.data_type is not None:
                    prop = (edge.label, node.value, node.data_type.value)
                else:
                    prop = (edge.label, node.value)
                sm[class_ids[edge.source_id]]['static_properties'].append(prop)

        for edge in drepr.sm.edges.values():
            if isinstance(drepr.sm.nodes[edge.source_id], ClassNode) and isinstance(
                    drepr.sm.nodes[edge.target_id], ClassNode):
                sm[class_ids[edge.source_id]]['links'].append((edge.label, class_ids[edge.target_id]))
            if edge.is_subject:
                sm[class_ids[edge.source_id]]['subject'] = drepr.sm.nodes[edge.target_id].attr_id

        sm['prefixes'] = drepr.sm.prefixes

        preprocessing: List[dict] = []
        for prepro in drepr.preprocessing:
            preprocessing.append(OrderedDict([("type", prepro.type.value)]))
            for k, v in asdict(prepro.value).items():
                preprocessing[-1][k] = v
            preprocessing[-1]["path"] = prepro.value.path.to_lang_format(use_json_path)

        return OrderedDict(
            [("version", version),
             ("resources",
              OrderedDict(
                  [(res.id,
                    OrderedDict([("type", res.type.value)] + (
                        [(k, v)
                         for k, v in asdict(res.prop).items()] if res.prop is not None else [])))
                   for res in drepr.resources])), ("preprocessing", preprocessing),
             ("attributes",
              OrderedDict([(attr.id,
                            OrderedDict([("resource_id", attr.resource_id),
                                         ("path", attr.path.to_lang_format(use_json_path)),
                                         ("unique", attr.unique), ("sorted", attr.sorted.value),
                                         ("value_type", attr.value_type.value),
                                         ("missing_values", attr.missing_values)]))
                           for attr in drepr.attrs])),
             ("alignments", [
                 OrderedDict([("type", AlignmentType.range.value), ("source", align.source),
                              ("target", align.target),
                              ("aligned_dims", [
                                  OrderedDict([
                                      ("source", step.source_idx),
                                      ("target", step.target_idx),
                                  ]) for step in align.aligned_steps
                              ])]) if isinstance(align, RangeAlignment) else
                 OrderedDict([("type", AlignmentType.value.value), ("source", align.source),
                              ("target", align.target)]) for align in drepr.aligns
             ]), ("semantic_model", sm)])


from dataclasses import dataclass
from enum import Enum
from typing import List, Dict, Optional, Union

from .attr import Attr


class DataType(Enum):
    xsd_decimal = "xsd:decimal"
    xsd_anyURI = "xsd:anyURI"
    xsd_gYear = "xsd:gYear"
    xsd_dateTime = "xsd:dateTime"
    xsd_int = "xsd:int"
    xsd_string = "xsd:string"


@dataclass
class ClassNode:
    node_id: str
    label: str

    def get_abs_iri(self, sm: 'SemanticModel'):
        """Get the absolute IRI of this node"""
        if sm.is_rel_iri(self.label):
            return sm.get_abs_iri(self.label)
        return self.label

    def get_rel_iri(self, sm: 'SemanticModel'):
        if sm.is_rel_iri(self.label):
            return self.label
        return sm.get_rel_iri(self.label)


@dataclass
class DataNode:
    node_id: str
    attr_id: str
    data_type: Optional[DataType] = None


@dataclass
class LiteralNode:
    node_id: str
    value: str
    data_type: Optional[DataType] = None


@dataclass
class Edge:
    edge_id: int
    source_id: str
    target_id: str
    label: str
    is_subject: bool = False
    is_required: bool = False

    def get_abs_iri(self, sm: 'SemanticModel'):
        """Get the absolute IRI of the predicate"""
        if sm.is_rel_iri(self.label):
            return sm.get_abs_iri(self.label)
        return self.label

    def get_rel_iri(self, sm: 'SemanticModel'):
        if sm.is_rel_iri(self.label):
            return self.label
        return sm.get_rel_iri(self.label)


Node = Union[LiteralNode, DataNode, ClassNode]


@dataclass
class SemanticModel:
    nodes: Dict[str, Node]
    edges: Dict[int, Edge]
    prefixes: Dict[str, str]

    @staticmethod
    def get_default(attrs: List[Attr]) -> 'SemanticModel':
        """
        Automatically generate a semantic model from a list of attributes.

        WARNING: the engine may not able to map data to this semantic model if the final output should be
        comprised of multiple tables.
        """
        prefixes = {"eg": "https://example.org/"}
        aids = {attr.id for attr in attrs}
        cid = None
        for i in range(len(attrs)):
            cid = f"c{i}"
            if cid not in aids:
                break
        nodes = {cid: ClassNode(cid, "eg:Record")}
        edges = {}
        for attr in attrs:
            nodes[attr.id] = DataNode(attr.id, attr.id, None)
            edge_id = len(edges)
            edges[edge_id] = Edge(edge_id, cid, attr.id, f"eg:{attr.id}")

        return SemanticModel(nodes, edges, prefixes)

    @staticmethod
    def get_default_prefixes() -> Dict[str, str]:
        return {
            'drepr': "https://purl.org/drepr/1.0/",
            'rdf': "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
            'rdfs': "http://www.w3.org/2000/01/rdf-schema#",
            'owl': "http://www.w3.org/2002/07/owl#"
        }

    @staticmethod
    def deserialize(raw: dict) -> "SemanticModel":
        nodes = {}
        for nid, n in raw['nodes'].items():
            if n['type'] == 'class_node':
                nodes[nid] = ClassNode(n['node_id'], n['label'])
            elif n['type'] == 'data_node':
                nodes[nid] = DataNode(n['node_id'], n['attr_id'], DataType(n['data_type']) if n['data_type'] is not None else None)
            elif n['type'] == 'literal_node':
                nodes[nid] = LiteralNode(n['node_id'], n['value'], DataType(n['data_type']) if n['data_type'] is not None else None)
            else:
                raise NotImplementedError()
        edges = {eid: Edge(**e) for eid, e in raw['edges'].items()}
        return SemanticModel(nodes, edges, raw['prefixes'])

    def remove_node(self, node_id: str):
        self.nodes.pop(node_id)
        removed_edges = []
        for eid, e in self.edges.items():
            if e.source_id == node_id or e.target_id == node_id:
                removed_edges.append(eid)
        for eid in removed_edges:
            self.edges.pop(eid)

    def class2dict(self, class_id: str) -> Dict[str, Union[List[int], int]]:
        """
        Get a dictionary that contains information (predicates) about a given class
        """
        info = {}
        for eid, e in self.edges.items():
            if e.source_id != class_id:
                continue

            if e.label in info:
                if not isinstance(info[e.label], list):
                    info[e.label] = [info[e.label], eid]
                else:
                    info[e.label].append(eid)
            else:
                info[e.label] = eid
        return info

    def iter_class_nodes(self):
        for n in self.nodes.values():
            if isinstance(n, ClassNode):
                yield n

    def iter_outgoing_edges(self, node_id: str):
        for e in self.edges.values():
            if e.source_id == node_id:
                yield e

    def iter_incoming_edges(self, node_id: str):
        for e in self.edges.values():
            if e.target_id == node_id:
                yield e

    def iter_child_nodes(self, node_id: str):
        for e in self.edges.values():
            if e.source_id == node_id:
                yield self.nodes[e.source_id]

    def iter_parent_nodes(self, node_id: str):
        for e in self.edges.values():
            if e.target_id == node_id:
                yield self.nodes[e.target_id]

    def get_rel_iri(self, abs_iri: str) -> str:
        """Convert an absolute IRI to a relative IRI. """
        assert not self.is_rel_iri(abs_iri)
        for prefix, uri in self.prefixes.items():
            if abs_iri.startswith(uri):
                return f"{prefix}:{abs_iri.replace(uri, '')}"
        raise ValueError("Create create relative IRI because there is no suitable prefix")

    def get_abs_iri(self, rel_iri: str) -> str:
        """Convert a relative IRI to an absolute IRI."""
        assert self.is_rel_iri(rel_iri)
        prefix, val = rel_iri.split(":", 1)
        if prefix not in self.prefixes:
            raise ValueError(f"Cannot create absolute IRI because the prefix {prefix} does not exist")
        return f"{self.prefixes[prefix]}{val}"

    def is_rel_iri(self, iri: str) -> bool:
        return iri.find("://") == -1 and iri.find(":") != -1



from collections import OrderedDict
from dataclasses import dataclass, asdict
from enum import Enum
from io import StringIO
from typing import List, Dict, Any, Optional

import ujson
from ruamel.yaml import YAML

from drepr.utils.validator import Validator, InputError
from .align import Alignment, RangeAlignment, AlignmentType, ValueAlignment, AlignedStep
from .attr import Attr
from .parse_v1 import ReprV1Parser
from .parse_v2 import ReprV2Parser
from .preprocessing import Preprocessing, PMap, PFilter, RMap, PSplit
from .resource import Resource, CSVProp
from .sm import SemanticModel, DataNode, ClassNode, LiteralNode

yaml = YAML()
yaml.Representer.add_representer(OrderedDict, yaml.Representer.represent_dict)


@dataclass
class EngineFormat:
    model: Dict[str, Any]
    edges_optional: List[bool]
    resource_idmap: Dict[str, int]
    attribute_idmap: Dict[str, int]
    sm_node_idmap: Dict[str, int]


@dataclass
class DRepr:
    resources: List[Resource]
    preprocessing: List[Preprocessing]
    attrs: List[Attr]
    aligns: List[Alignment]
    sm: SemanticModel

    @staticmethod
    def parse(raw: dict) -> "DRepr":
        Validator.must_have(raw, 'version', "Parsing D-REPR configuration")
        if raw['version'] == '1':
            model = ReprV1Parser.parse(raw)
            model.is_valid()
            return model
        elif raw['version'] == '2':
            model = ReprV2Parser.parse(raw)
            model.is_valid()
            return model
        raise InputError(f"Parsing error, get unknown version: {raw['version']}")

    @staticmethod
    def parse_from_file(fpath: str) -> "DRepr":
        if fpath.endswith(".json"):
            with open(fpath, "r") as f:
                return DRepr.parse(ujson.load(f))

        if fpath.endswith(".yml") or fpath.endswith(".yaml"):
            with open(fpath, 'r') as f:
                return DRepr.parse(yaml.load(f))

        raise Exception(f"Does not supported this file: {fpath}. Only support json or yaml file")

    @staticmethod
    def empty() -> "DRepr":
        return DRepr([], [], [], [], SemanticModel({}, [], {}))

    @staticmethod
    def deserialize(raw: dict) -> "DRepr":
        resources = [Resource.deserialize(o) for o in raw['resources']]
        preprocessing = [Preprocessing.deserialize(o) for o in raw['preprocessing']]
        attrs = [Attr.deserialize(o) for o in raw['attrs']]
        aligns = []

        for align in raw['aligns']:
            if align['type'] == AlignmentType.range.value:
                aligns.append(
                    RangeAlignment(align['source'], align['target'], [
                        AlignedStep(step['source_idx'], step['target_idx'])
                        for step in align['aligned_steps']
                    ]))
            elif align['type'] == AlignmentType.value.value:
                aligns.append(ValueAlignment(align['source'], align['target']))
            else:
                raise NotImplementedError()
        sm = SemanticModel.deserialize(raw['sm'])

        return DRepr(resources, preprocessing, attrs, aligns, sm)

    def serialize(self) -> dict:
        obj = asdict(self)
        # post-process the enum
        for res in obj['resources']:
            res['type'] = res['type'].value
        for prepro in obj['preprocessing']:
            prepro['type'] = prepro['type'].value
            for i, step in enumerate(prepro['value']['path']['steps']):
                if isinstance(step, Enum):
                    prepro['value']['path']['steps'][i] = step.value

        for attr in obj['attrs']:
            attr['sorted'] = attr['sorted'].value
            for i, step in enumerate(attr['path']['steps']):
                if isinstance(step, Enum):
                    attr['steps'][i] = step.value

            attr['value_type'] = attr['value_type'].value
        for node in obj['sm']['nodes'].values():
            if node.get('data_type', None) is not None:
                node['data_type'] = node['data_type'].value

        # adding a bit of meta-data about the alignment
        for align, raw_align in zip(self.aligns, obj['aligns']):
            if isinstance(align, RangeAlignment):
                raw_align['type'] = AlignmentType.range.value
            elif isinstance(align, ValueAlignment):
                raw_align['type'] = AlignmentType.value.value
            else:
                raise NotImplementedError()

        # similarly, add meta-data about the nodes
        for node in obj['sm']['nodes'].values():
            if isinstance(self.sm.nodes[node['node_id']], ClassNode):
                node['type'] = 'class_node'
            elif isinstance(self.sm.nodes[node['node_id']], DataNode):
                node['type'] = 'data_node'
            elif isinstance(self.sm.nodes[node['node_id']], LiteralNode):
                node['type'] = 'literal_node'
            else:
                raise NotImplementedError()

        return obj

    def is_valid(self):
        """
        Perform a check to see if this D-REPR is valid. Raise AssertionError if this is not valid
        """
        # CHECK 1: all references (resource id, attribute ids) are valid
        resource_ids = {r.id for r in self.resources}
        for pref in self.preprocessing:
            if pref.value.output is not None:
                # preprocessing create new resource
                assert pref.value.output not in resource_ids, "Cannot overwrite existing resources"
                resource_ids.add(pref.value.output)
        attr_ids = {attr.id for attr in self.attrs}

        for attr in self.attrs:
            assert attr.resource_id in resource_ids, f"Attribute {attr.resource_id} does not belong to any resources"
        for align in self.aligns:
            assert align.source in attr_ids and align.target in attr_ids, \
                f"The alignment {align} links to non-existence attributes"

        for node in self.sm.nodes.values():
            if isinstance(node, DataNode):
                assert node.attr_id in attr_ids, f"The semantic model has a link to " \
                                                 f"a non-existence attribute: {node.attr_id}"

        # CHECK 2: check class and predicates are valid
        for node in self.sm.nodes.values():
            if isinstance(node, ClassNode):
                if self.sm.is_rel_iri(node.label):
                    prefix = node.label.split(":", 1)[0]
                    assert prefix in self.sm.prefixes, f"Unknown prefix `{prefix}` of the " \
                                                       f"ontology class {node.label}"
        for edge in self.sm.edges.values():
            if self.sm.is_rel_iri(edge.label):
                prefix = edge.label.split(":", 1)[0]
                assert prefix in self.sm.prefixes, f"Unknown prefix `{prefix}` of the " \
                                                   f"ontology predicate {edge.label}"

    def to_lang_format(self, simplify: bool = True, use_json_path: bool = False) -> dict:
        return ReprV2Parser.dump(self, simplify, use_json_path)

    def to_lang_yml(self, simplify: bool = True, use_json_path: bool = False) -> str:
        model = self.to_lang_format(simplify, use_json_path)
        out = StringIO()
        yaml.dump(model, out)
        return out.getvalue()

    def to_engine_format(self) -> EngineFormat:
        """
        Turn this D-REPR configuration into the format that the engine can read
        :return:
        """
        # map string id to incremental numbers (for resource id and attribute id)
        ridmap = OrderedDict()
        for resource in self.resources:
            ridmap[resource.id] = len(ridmap)
        for pref in self.preprocessing:
            if pref.value.output is not None:
                ridmap[pref.value.output] = len(ridmap)
        aidmap = OrderedDict()
        for attr in self.attrs:
            aidmap[attr.id] = len(aidmap)

        resources = []
        for res in self.resources:
            resources.append({"type": res.type.value})
            if isinstance(res.prop, CSVProp):
                resources[-1]['value'] = {
                    "resource_id": ridmap[res.id],
                    "delimiter": res.prop.delimiter
                }
            else:
                resources[-1]['value'] = ridmap[res.id]

        preprocessing = []
        for pref in self.preprocessing:
            prepro = {
                "type": pref.type.value,
                "resource_id": ridmap[pref.value.resource_id],
                "path": pref.value.path.to_engine_format(),
                "output": ridmap[pref.value.output] if pref.value.output is not None else None,
            }
            if isinstance(pref.value, PMap):
                prepro['code'] = pref.value.code
                prepro['change_structure'] = pref.value.change_structure
            elif isinstance(pref.value, (PFilter, PSplit)):
                prepro['code'] = pref.value.code
            elif isinstance(pref.value, RMap):
                prepro['func_id'] = {"t": pref.value.func_id.value}
            else:
                raise NotImplementedError()
            preprocessing.append(prepro)

        attributes = [{
            "id": aidmap[a.id],
            "resource_id": ridmap[a.resource_id],
            "path": a.path.to_engine_format(),
            "unique": a.unique,
            "sorted": a.sorted.value,
            "vtype": a.value_type.value,
            "missing_values": [self._serde_engine_value(v) for v in a.missing_values]
        } for a in self.attrs]

        alignments = []
        for align in self.aligns:
            if isinstance(align, RangeAlignment):
                alignments.append({
                    "type": AlignmentType.range.value,
                    "source": aidmap[align.source],
                    "target": aidmap[align.target],
                    "aligned_dims": [{
                        "source": ad.source_idx,
                        "target": ad.target_idx
                    } for ad in align.aligned_steps]
                })
            elif isinstance(align, ValueAlignment):
                alignments.append({
                    "type": AlignmentType.value.value,
                    "source": aidmap[align.source],
                    "target": aidmap[align.target]
                })
            else:
                raise NotImplementedError()

        engine_sm: Dict[str, Any] = {
            "nodes": [],
            "edges": [],
            "prefixes": [(prefix, uri) for prefix, uri in self.sm.prefixes.items()]
        }
        nodes = {"cnodes": [], "dnodes": [], "lnodes": []}
        for node_id, node in self.sm.nodes.items():
            if isinstance(node, ClassNode):
                nodes['cnodes'].append({
                    "type": "class_node",
                    "node_id": node_id,
                    "rel_label": node.label,
                    "abs_label": self.sm.get_abs_iri(node.label)
                })
            elif isinstance(node, DataNode):
                nodes['dnodes'].append({
                    "type": "data_node",
                    "node_id": node_id,
                    "attr_id": aidmap[node.attr_id],
                    "data_type": node.data_type.value if node.data_type is not None else None,
                })
            elif isinstance(node, LiteralNode):
                nodes["lnodes"].append({
                    "type": "literal_node",
                    "node_id": node_id,
                    "val": self._serde_engine_value(node.value),
                    "data_type": node.data_type.value if node.data_type is not None else None,
                })
            else:
                raise NotImplementedError()

        for k in ["cnodes", "dnodes", "lnodes"]:
            for n in nodes[k]:
                engine_sm['nodes'].append(n)
        nidmap = {}
        for i, n in enumerate(engine_sm["nodes"]):
            nidmap[n['node_id']] = i
            n['node_id'] = i

        for eid, edge in self.sm.edges.items():
            engine_sm['edges'].append({
                "edge_id": eid,
                "source": nidmap[edge.source_id],
                "target": nidmap[edge.target_id],
                "rel_label": edge.label,
                "abs_label": self.sm.get_abs_iri(edge.label),
                "is_subject": edge.is_subject
            })

        edges_optional = [
            not edge.is_required
            for edge in self.sm.edges.values()
        ]

        return EngineFormat({
            "resources": resources,
            "preprocessing": preprocessing,
            "attributes": attributes,
            "alignments": alignments,
            "semantic_model": engine_sm
        }, edges_optional, ridmap, aidmap, nidmap)

    def _serde_engine_value(self, value: Any):
        """Serialize a python value to a json representation of the Value struct in the Rust engine"""
        if value is None:
            return {"t": "Null"}
        elif isinstance(value, bool):
            return {"t": "Bool", "c": value}
        elif isinstance(value, int):
            return {"t": "I64", "c": value}
        elif isinstance(value, float):
            return {"t": "F64", "c": value}
        elif isinstance(value, str):
            return {"t": "Str", "c": value}
        elif isinstance(value, list):
            return {"t": "Array", "c": [self._serde_engine_value(v) for v in value]}
        elif isinstance(value, (dict, OrderedDict)):
            return {"t": "Object", "c": {
                k: self._serde_engine_value(v)
                for k, v in value.items()
            }}
        else:
            raise InputError(f"Cannot serialize the value of type: {type(value)} to JSON")

    def remove_resource(self, resource_id: str):
        self.resources = [r for r in self.resources if r.id != resource_id]
        for i in range(len(self.preprocessing) - 1, -1, -1):
            if self.preprocessing[i].value.resource_id == resource_id:
                self.preprocessing.pop(i)

        for i in range(len(self.attrs) - 1, -1, -1):
            if self.attrs[i].resource_id == resource_id:
                self.remove_attribute(self.attrs[i].id, idx=i)

    def get_resource_by_id(self, resource_id: str) -> Optional[Resource]:
        for r in self.resources:
            if r.id == resource_id:
                return r
        return None

    def get_attr_by_id(self, attr_id: str) -> Optional[Attr]:
        for a in self.attrs:
            if a.id == attr_id:
                return a
        return None

    def remove_attribute(self, attr_id: str, idx: Optional[int] = None):
        if idx is None:
            idx = next(i for i in range(len(self.attrs), -1, -1) if self.attrs[i].id == attr_id)

        self.attrs.pop(idx)
        for i in range(len(self.aligns) - 1, -1, -1):
            if self.aligns[i].source == attr_id or self.aligns[i].target == attr_id:
                self.aligns.pop(i)

        for node in self.sm.nodes:
            if isinstance(node, DataNode) and node.attr_id == attr_id:
                self.sm.remove_node(node.node_id)

    def update_attribute(self, attr_id: str, new_attr: Attr):
        for i, attr in enumerate(self.attrs):
            if attr.id == attr_id:
                self.attrs[i] = new_attr

        for align in self.aligns:
            if align.source == attr_id:
                align.source = new_attr.id
            elif align.target == attr_id:
                align.target = new_attr.id

        for node in self.sm.nodes:
            if isinstance(node, DataNode) and node.attr_id == attr_id:
                node.attr_id = new_attr.id



class PyExec:
    def __init__(self, func):
        self.func = func

    @staticmethod
    def compile(code: str) -> "PyExec":
        lines = code.split("\n")
        indent = PyExec.detect_indent(lines)

        code = "def func(value, index, context):\n" + "\n".join([
            indent + line
            for line in lines
        ])
        session_locals = {}
        exec(code, None, session_locals)
        return PyExec(session_locals['func'])

    def exec(self, value, index, context):
        return self.func(value, index, context)

    @staticmethod
    def detect_indent(lines: str) -> str:
        indent = "\t"
        for line in lines:
            if line.startswith("\t"):
                break

            if line.startswith(" "):
                n = 0
                for c in line:
                    if c != " ":
                        indent = " " * n
                        break
                    n += 1
                break
        return indent



from abc import ABC, abstractmethod
from typing import Union, List

from drepr.models import IndexExpr, RangeExpr

Index = Union[str, int]


class NDArrayReader(ABC):

    @abstractmethod
    def get_value(self, index: List[Index]):
        """
        Get value at the given index
        """
        pass

    @abstractmethod
    def set_value(self, index: List[Index], value):
        """
        Assigning value to the given index
        """
        pass

    @abstractmethod
    def select(self, steps: List[Union[IndexExpr, RangeExpr]]):
        """
        Select values that match the path.
        """
        pass

    @abstractmethod
    def len(self) -> int:
        """
        Get length of the current node.
        """
        pass

    @abstractmethod
    def len_range(self) -> int:
        """
        Get length of the current node (range only).
        """
        pass



#!/usr/bin/python
# -*- coding: utf-8 -*-

import argparse
import time
from pathlib import Path

import ujson
from ruamel.yaml import YAML

from drepr.engine import execute, FileOutput, OutputFormat, MemoryOutput
from drepr.models import DRepr

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-r",
        "--repr",
        required=True,
        help="either a string containing representation of a dataset (JSON format) "
             "or a path to a file containing representation (support 2 formats: JSON & YML)")

    parser.add_argument("-d",
                        "--resource",
                        required=True,
                        nargs="+",
                        help="file paths of resources in this format: <resource_id>=<file_path>;")
    parser.add_argument("-v",
                        "--verbose",
                        default=0,
                        help="increase output verbosity",
                        action="count")
    parser.add_argument("-o", "--output", help="the file we will write output to")
    parser.add_argument("-f", "--format",
                        choices=[e.value for e in OutputFormat],
                        default=OutputFormat.TTL, help="format of the data (default is ttl)")
    args = parser.parse_args()

    # read repr
    try:
        raw = ujson.loads(args.repr)
        ds_model = DRepr.parse(raw)
    except ValueError:
        if not Path(args.repr).exists():
            print(f"Cannot read the representation file at: {args.repr}")
            exit(1)
        ds_model = None

    if ds_model is None:
        ds_model = DRepr.parse_from_file(args.repr)

        if args.repr.endswith(".json"):
            with open(args.repr, "r") as f:
                raw_repr = ujson.load(f)
        elif args.repr.endswith(".yml") or args.repr.endswith(".yaml"):
            yaml = YAML()
            with open(args.repr, "r") as f:
                raw_repr = yaml.load(f)
        else:
            print(
                f"Cannot determine the format of the representation file. Should ends with .json or .yml"
            )
            exit(1)

    # read the resources
    resources = {}
    for r in args.resource:
        if r.find("=") == -1:
            print(f"Invalid resource format. Expect `=`")
            parser.print_help()
            exit(1)
        resource_id, resource_path = r.split("=")
        if not Path(resource_path).exists():
            print(f"Resource file: {resource_path} does not exist")
            exit(1)
        resources[resource_id] = resource_path

    # map data to RDF
    if args.output is not None:
        output = FileOutput(args.output, OutputFormat(args.format))
    else:
        output = MemoryOutput(OutputFormat(args.format))

    start = time.time()
    result = execute(ds_model, resources, output, args.verbose > 1)
    end = time.time()

    if args.verbose > 0:
        print(">>> [DREPR] runtime: %.5f seconds" % (end - start))

    if args.output is None:
        print(">>> Dumping RDF triples to console...\n")
        print(result['value'])



from typing import NamedTuple, Union

ResourceDataFile = NamedTuple("ResourceDataType", [("file_path", str)])
ResourceDataString = NamedTuple("ResourceDataString", [("value", bytes)])
ResourceData = Union[ResourceDataString, ResourceDataFile]

from .nc_patch import *
from .jp_propname_patch import *
from .xml_patch import *



import copy
import re

from drepr.utils.validator import Validator, InputError
from ..sm import SemanticModel, ClassNode, DataNode, Edge, LiteralNode, DataType


class SMParser:
    """
    SM has the following schema

    ```
    semantic_model:
      <class_id>:
        properties:
            - [<predicate>, <attr_id>, <sem_type>, <is_required=false>]
        links:
            - [<predicate>, <value>, <sem_type>]
        static_properties:
            -
      prefixes:
        <prefix>: <uri>
    ```
    """
    CLS_KEYS = {"properties", "subject", "links", "static_properties"}
    DATA_TYPE_VALUES = {x.value for x in DataType}

    REG_SM_CLASS = re.compile(r"^((.+):\d+)$")
    REG_SM_DNODE = re.compile(r"^((?:(?!--).)+:\d+)--((?:(?!\^\^).)+)(?:\^\^(.+))?$")
    REG_SM_LNODE = re.compile(
        r"^((?:(?!--).)+:\d+)--((?:(?!--).)+)--((?:(?!\^\^).)+)(?:\^\^(.+))?$")
    REG_SM_REL = re.compile(r"^((?:(?!--).)+:\d+)--((?:(?!--).)+)--((?:(?!--).)+:\d+)$")

    @classmethod
    def parse(cls, sm: dict) -> SemanticModel:
        nodes = {}
        edges = {}
        # shallow copy
        sm = copy.copy(sm)
        prefixes = sm.pop('prefixes', {})

        trace0 = f"Parsing `prefixes` of the semantic model"
        Validator.must_be_dict(prefixes, trace0)
        for prefix, uri in prefixes.items():
            Validator.must_be_str(uri, f"{trace0}\nParse prefix {prefix}")

        for class_id, class_conf in sm.items():
            trace0 = f"Parsing class `{class_id}` of the semantic model"
            Validator.must_be_dict(class_conf, trace0)
            Validator.must_be_subset(cls.CLS_KEYS, class_conf.keys(), "keys of ontology class",
                                     trace0)

            try:
                class_name = cls.REG_SM_CLASS.match(class_id).group(2)
            except Exception as e:
                raise InputError(
                    f"{trace0}\nERROR: invalid class_id `{class_id}`. Expect to be <string>:<number>"
                )

            nodes[class_id] = ClassNode(class_id, class_name)

        for class_id, class_conf in sm.items():
            trace0 = f"Parsing class `{class_id}` of the semantic model"

            for i, prop in enumerate(class_conf.get('properties', [])):
                trace1 = f"{trace0}\nParsing property {i}: {prop}"
                if len(prop) == 2:
                    predicate, attr_id = prop
                    data_type = None
                    is_required = False
                elif len(prop) == 3:
                    predicate, attr_id, data_type = prop
                    if isinstance(data_type, bool) or data_type.lower() in {"true", "false"}:
                        is_required = data_type if isinstance(data_type,
                                                              bool) else data_type == "true"
                        data_type = None
                    else:
                        is_required = False
                elif len(prop) == 4:
                    predicate, attr_id, data_type, is_required = prop
                else:
                    raise InputError(
                        f"{trace1}\nERROR: Expect value of the property to be an array of two "
                        f"three or four items (<predicate>, <attribute_id>[, semantic_type='auto'][, is_required=false])"
                    )

                if data_type is not None:
                    Validator.must_in(data_type, cls.DATA_TYPE_VALUES,
                                      f"{trace1}\nParsing data type")
                    data_type = DataType(data_type)

                node = DataNode(node_id=f"dnode:{attr_id}", attr_id=attr_id, data_type=data_type)
                nodes[node.node_id] = node
                edges[len(edges)] = Edge(len(edges), class_id, node.node_id, predicate, is_required=is_required)

            for i, link_conf in enumerate(class_conf.get('links', [])):
                trace1 = f"{trace0}\nParsing link {i}: {link_conf}"
                if len(link_conf) != 2:
                    raise InputError(
                        f"{trace1}\nERROR: Expect value of the link to be an array of two "
                        f"items (<predicate>, <class_id>)")
                predicate, object_class_id = link_conf
                edges[len(edges)] = Edge(len(edges), class_id, object_class_id, predicate)

            for i, prop in enumerate(class_conf.get('static_properties', [])):
                trace1 = f"{trace0}\nParsing static properties {i}: {prop}"
                if len(prop) == 2:
                    predicate, value = prop
                    data_type = None
                elif len(prop) == 3:
                    predicate, value, data_type = prop
                else:
                    raise InputError(
                        f"{trace1}\nERROR: Expect value of the property to be an array of two "
                        f"or three items (<predicate>, <attribute_id>, [semantic_type])")

                if data_type is not None:
                    Validator.must_in(data_type, cls.DATA_TYPE_VALUES,
                                      f"{trace1}\nParsing data type")
                    data_type = DataType(data_type)

                node = LiteralNode(node_id=f"lnode:{len(nodes)}", value=value, data_type=data_type)
                nodes[node.node_id] = node
                edges[len(edges)] = Edge(len(edges), class_id, node.node_id, predicate)

        for class_id, class_conf in sm.items():
            trace0 = f"Parsing class `{class_id}` of the semantic model"
            if 'subject' in class_conf:
                trace1 = f"{trace0}\nParsing subject"
                attr_id = class_conf['subject']
                target_id = f"dnode:{attr_id}"
                for edge in edges.values():
                    if edge.source_id == class_id and edge.target_id == target_id:
                        edge.is_subject = True
                        break
                else:
                    raise InputError(f"{trace1}\nERROR: Subject of the class node must be one "
                                     f"of the attributes used in the semantic model")

        return SemanticModel(nodes, edges, prefixes)
